@@ -1,5 +1,8 @@
 {-# LANGUAGE CPP #-}
+
+#if !(MIN_VERSION_base(4,16,0)) || !(MIN_VERSION_transformers(0,6,0))
 {-# OPTIONS_GHC -fno-warn-deprecations #-}
+#endif
 
 #define GHC_GENERICS_OK __GLASGOW_HASKELL__ >= 702
 
@@ -39,7 +42,11 @@ module Data.Functor.Invariant
 #endif
   , WrappedFunctor(..)
   , invmapContravariant
+  , invmapProfunctor
+  , invmapArrow
   , WrappedContravariant(..)
+  , InvariantProfunctor(..)
+  , InvariantArrow(..)
     -- * @Invariant2@
   , Invariant2(..)
   , invmap2Bifunctor
@@ -75,7 +82,10 @@ import           Data.Monoid (Alt(..))
 #endif
 import           Data.Monoid (Dual(..), Endo(..))
 import           Data.Proxy (Proxy(..))
-import qualified Data.Semigroup as Semigroup (First(..), Last(..), Option(..))
+import qualified Data.Semigroup as Semigroup (First(..), Last(..))
+#if !(MIN_VERSION_base(4,16,0))
+import qualified Data.Semigroup as Semigroup (Option(..))
+#endif
 import           Data.Semigroup (Min(..), Max(..), Arg(..))
 import qualified Data.Traversable as T (Traversable(..))
 #if GHC_GENERICS_OK
@@ -143,10 +153,8 @@ import           Data.Tagged (Tagged(..))
 import           Control.Applicative.Backwards (Backwards(..))
 import           Control.Applicative.Lift (Lift(..))
 import           Control.Monad.Trans.Cont (ContT)
-import           Control.Monad.Trans.Error (ErrorT(..))
 import           Control.Monad.Trans.Except (ExceptT(..), runExceptT)
 import           Control.Monad.Trans.Identity (IdentityT, mapIdentityT)
-import           Control.Monad.Trans.List (ListT, mapListT)
 import           Control.Monad.Trans.Maybe (MaybeT, mapMaybeT)
 import qualified Control.Monad.Trans.RWS.Lazy as Lazy (RWST(..))
 import qualified Control.Monad.Trans.RWS.Strict as Strict (RWST(..))
@@ -155,6 +163,10 @@ import qualified Control.Monad.Trans.State.Lazy as Lazy (StateT(..))
 import qualified Control.Monad.Trans.State.Strict as Strict (StateT(..))
 import qualified Control.Monad.Trans.Writer.Lazy as Lazy (WriterT, mapWriterT)
 import qualified Control.Monad.Trans.Writer.Strict as Strict (WriterT, mapWriterT)
+#if !(MIN_VERSION_transformers(0,6,0))
+import           Control.Monad.Trans.Error (ErrorT(..))
+import           Control.Monad.Trans.List (ListT, mapListT)
+#endif
 import           Data.Functor.Constant (Constant(..))
 import           Data.Functor.Reverse (Reverse(..))
 
@@ -187,6 +199,14 @@ invmapFunctor = flip $ const fmap
 invmapContravariant :: Contravariant f => (a -> b) -> (b -> a) -> f a -> f b
 invmapContravariant = const contramap
 
+-- | A 'Profunctor' with the same input and output types can be seen as an 'Invariant' functor.
+invmapProfunctor :: Profunctor p => (a -> b) -> (b -> a) -> p a a -> p b b
+invmapProfunctor = flip dimap
+
+-- | An 'Arrow' with the same input and output types can be seen as an 'Invariant' functor.
+invmapArrow :: Arrow arr => (a -> b) -> (b -> a) -> arr a a -> arr b b
+invmapArrow fn1 fn2 arrow = arr fn1 Cat.. arrow Cat.. arr fn2
+
 -------------------------------------------------------------------------------
 -- Invariant instances
 -------------------------------------------------------------------------------
@@ -303,11 +323,13 @@ instance Invariant Semigroup.First where
 instance Invariant Semigroup.Last where
   invmap = invmapFunctor
 -- | from "Data.Semigroup"
-instance Invariant Semigroup.Option where
-  invmap = invmapFunctor
--- | from "Data.Semigroup"
 instance Invariant (Arg a) where
   invmap = invmapFunctor
+#if !(MIN_VERSION_base(4,16,0))
+-- | from "Data.Semigroup"
+instance Invariant Semigroup.Option where
+  invmap = invmapFunctor
+#endif
 
 -- | from "System.Console.GetOpt"
 instance Invariant ArgDescr where
@@ -472,9 +494,6 @@ instance Invariant f => Invariant (Lift f) where
 -- | from the @transformers@ package
 instance Invariant (ContT r m) where
   invmap = invmapFunctor
--- -- | from the @transformers@ package
-instance Invariant m => Invariant (ErrorT e m) where
-  invmap f g = ErrorT . invmap (invmap f g) (invmap g f) . runErrorT
 -- | from the @transformers@ package
 instance Invariant m => Invariant (ExceptT e m) where
   invmap f g = ExceptT . invmap (invmap f g) (invmap g f) . runExceptT
@@ -482,9 +501,6 @@ instance Invariant m => Invariant (ExceptT e m) where
 instance Invariant m => Invariant (IdentityT m) where
   invmap f g = mapIdentityT (invmap f g)
 -- | from the @transformers@ package
-instance Invariant m => Invariant (ListT m) where
-  invmap f g = mapListT $ invmap (invmap f g) (invmap g f)
--- | from the @transformers@ package
 instance Invariant m => Invariant (MaybeT m) where
   invmap f g = mapMaybeT $ invmap (invmap f g) (invmap g f)
 -- | from the @transformers@ package
@@ -530,6 +546,14 @@ instance Invariant (Constant a) where
 -- | from the @transformers@ package
 instance Invariant f => Invariant (Reverse f) where
   invmap f g (Reverse a) = Reverse (invmap f g a)
+#if !(MIN_VERSION_transformers(0,6,0))
+-- | from the @transformers@ package
+instance Invariant m => Invariant (ErrorT e m) where
+  invmap f g = ErrorT . invmap (invmap f g) (invmap g f) . runErrorT
+-- | from the @transformers@ package
+instance Invariant m => Invariant (ListT m) where
+  invmap f g = mapListT $ invmap (invmap f g) (invmap g f)
+#endif
 
 -- | from the @unordered-containers@ package
 instance Invariant (HashMap k) where
@@ -563,9 +587,11 @@ instance Alternative f => Alternative (WrappedFunctor f) where
   many = WrapFunctor . many . unwrapFunctor
 
 instance Monad m => Monad (WrappedFunctor m) where
-  return = WrapFunctor . return
   WrapFunctor x >>= f = WrapFunctor (x >>= unwrapFunctor . f)
+#if !(MIN_VERSION_base(4,11,0))
+  return = WrapFunctor . return
   WrapFunctor a >> WrapFunctor b = WrapFunctor (a >> b)
+#endif
 
 instance MonadPlus m => MonadPlus (WrappedFunctor m) where
   mzero = WrapFunctor mzero
@@ -972,3 +998,19 @@ these, you can derive them using the "Data.Functor.Invariant.TH" module.
 genericInvmap :: (Generic1 f, Invariant (Rep1 f)) => (a -> b) -> (b -> a) -> f a -> f b
 genericInvmap f g = to1 . invmap f g . from1
 #endif
+
+-------------------------------------------------------------------------------
+-- Wrappers
+-------------------------------------------------------------------------------
+
+-- | A 'Profunctor' with the same input and output types can be seen as an 'Invariant' functor.
+newtype InvariantProfunctor p a = InvariantProfunctor (p a a)
+
+instance Profunctor p => Invariant (InvariantProfunctor p) where
+  invmap fn1 fn2 (InvariantProfunctor f) = InvariantProfunctor (invmapProfunctor fn1 fn2 f)
+
+-- | An 'Arrow' with the same input and output types can be seen as an 'Invariant' functor.
+newtype InvariantArrow c a = InvariantArrow (c a a)
+
+instance Arrow c => Invariant (InvariantArrow c) where
+  invmap fn1 fn2 (InvariantArrow arrow) = InvariantArrow (invmapArrow fn1 fn2 arrow)
